/* $Header: /cvsroot/sid/sid/user/c_tsw.h,v 1.20 2004/04/17 14:10:32 deppeler Exp $ */

#ifndef _C_TSW_H
#define _C_TSW_H

#include <stdio.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <strings.h>
#include <string.h>
#include <libgen.h>
#include <stdlib.h>
#include <time.h>

#ifndef TRUE
  #define FALSE	0
  #define TRUE	( !FALSE )
#endif /* TRUE */

/* Size of atomic reads.  */
#ifndef BUFSIZ
# define BUFSIZ (512 * 8)
#endif

#ifndef MIN
# define MIN(a,b) (((a) < (b)) ? (a) : (b))
#endif

#ifndef CHAR_BIT
# define CHAR_BIT 8
#endif

#define TYPE_MINIMUM(t) ((t) (TYPE_SIGNED (t) \
                              ? ~ (t) 0 << (sizeof (t) * CHAR_BIT - 1) : (t) 0))
#define TYPE_MAXIMUM(t) ((t) (~ (t) 0 - TYPE_MINIMUM (t)))

#ifndef OFF_T_MAX
# define OFF_T_MAX TYPE_MAXIMUM (off_t)
#endif

#define COPY_TO_EOF OFF_T_MAX


#define MAX_LINE_LENGTH 256
#define SHA1_HASH_SIZE 20
#define MAX_COMMANDS 256
#define NR_OF_WEEKDAYS 7
#define HASH_LENGTH 45	    		/* hash + spaces + '\0' */
#define HASH_ANNOUNCEMENT_LENGTH 7	/* digest= */
#define TTY_ANNOUNCEMENT_LENGTH 4	/* tty= */
#define TTY_MIN_LENGTH 1
#define MIN_DIGEST_LINE_LENGTH 70	/* see comment below */
#define MATCH_OPEN "sid: open for terminal device "
#define FMT_OPEN "sid: open for terminal device %s"
#define FMT_MATCH_HOST "%*s %*d %*d:%*d:%*d %s "
#define FMT_MATCH_TSW " tsw: tty=%d line=%d digest=%8[0123456789abcdef] %8[0123456789abcdef] %8[0123456789abcdef] %8[0123456789abcdef] %8[0123456789abcdef]"
#define MATCH_CLOSE "sid: close for terminal device "
#define FMT_CLOSE "sid: close for terminal device %s"
#define WEEK_SECONDS 604800 		/* (7*24*60*60) */
#define DAY_SECONDS  86400		/* (24*60*60) */
#define HOUR_SECONDS 3600		/* (60*60) */
#define MINUTE_SECONDS 60

/*
                                            70        60        50    44  40        30        20        10
    09876543210987654321098765432109876543210987654321098765432109876543210987654321098765432109876543210987654321
Apr  2 21:07:15 blade tsw: [ID 769231 kern.notice] tsw: tty=11 digest=e2cef1f8 dcab8cbc 6d6c7c74 06a22505 25b7be9f
                                                               1234567
Feb 23 15:48:12 debianvanilla kernel: tsw: tty=0 line=387 digest=8a90d28c 8825412f 6008da90 f8770c2b 05282358
 */

enum {
	NONE = 10000,
	INTERACTIVE = 10001,
	PASSWORD = 10002,
	SET = 10003,
	UNSET = 10004,
	PERIODIC = 10005,
	SINGULAR = 10006,
	WITHIN_HOST_DEF = 10007,    /* warning: also defined in lexer */
	WITHIN_COMMAND_DEF = 10008,  /* warning: also defined in lexer */
	WITHIN_GLOBAL_DEF = 10009,  /* warning: also defined in lexer */
       	GLOBAL = 10010,
	NORMAL_CMD = 10011,
	ANY_CMD = 10012,
	SHA1 = 10013
};


/*******************/
/* data structures */
/*******************/

/* tsw_time_t holds a list of time specifications */
typedef struct tsw_time_t { 
	int 			type;	/* periodic (weekday) or singular */
	int 			weekday[NR_OF_WEEKDAYS];
	time_t 			ctime;  /* point in time start time (unused) */
	int 			start_hour;
	int 			start_minute;
	int 			cont_day;
	int 			cont_hour;
	int 			cont_minute;
	int 			cont_second;
	struct se_time_t	*se;    /* once calculated start and end times
(from time specs above) to faciliate time checking at run time */
	struct tsw_time_t 	*next;
} tsw_time_t;

/* hour minute second */
typedef struct hms_t {
	int hour;
	int minute;
	int second;
} hms_t;

/* se_time_t holds a list of start and end time specs 
   (these are calculated from the specs in tsw_time_t)
 */
typedef struct se_time_t {
	int 			weekday[NR_OF_WEEKDAYS];
	hms_t			start;
	hms_t			end;
	struct se_time_t	*next;
} se_time_t;

typedef struct file_spec
{
        int 	fd;
	off_t 	size;
} file_spec;

/* tsw_cmt_t holds a list of command strings and their corresponding digests */
typedef struct tsw_cmd_t {
	char 		command[MAX_LINE_LENGTH]; /* read in */
	unsigned char	digest[SHA1_HASH_SIZE]; /* calculated value */
	int		mode;	    /* type of command: specified string or an
'any' type command */
	struct tsw_cmd_t *prev;
	struct tsw_cmd_t *next;
} tsw_cmd_t;

/* cmd_def_t holds a list of command definitions */
typedef struct cmd_def_t {
	char	name[MAX_LINE_LENGTH]; 
	struct 	tsw_time_t *time; /* when these commands are valid */
	short	ordered; 	  /* whether they must appear in order */
	tsw_cmd_t *command_list;  /* commands and their hashed values */
	struct  cmd_def_t *next;
} cmd_def_t;

/* cmd_info_t holds a pair of pointers to a command definition and a command
 * string/digest
 * used a helper structure to return pairs of pointers */
typedef struct cmd_info_t cmd_info_t;
struct cmd_info_t {
	cmd_def_t	*cmd_def;
	tsw_cmd_t	*command;
};

/* host_cft_t holds a list of host definitions */
typedef struct host_cfg_t {
	char	name[MAX_LINE_LENGTH]; /* hostname (as it appears in logfile */
	short  	admin_mode;	       /* administration mode (overrides global) */ 
	unsigned char	admin_digest[SHA1_HASH_SIZE]; /* password hash (calculated) */
	int	max_failures; /* TODO: number of trials for admin mode */
	struct 	tsw_time_t *automated_time; /* whence are automated actions */
	char	automated_action[MAX_LINE_LENGTH]; /* full command line */
	char	syslog_message[MAX_LINE_LENGTH];   /* intrusion log alert */
	char	syslog_facility[MAX_LINE_LENGTH];  /* facility for the above */
	int	facility;
	char	syslog_priority[MAX_LINE_LENGTH];  /* priority for the above */
	int	priority;
	char	syslog_tag[MAX_LINE_LENGTH];       /* 'tag' for the above */
	struct 	tsw_cmd_t *allowed_command_set; 
	struct	host_cfg_t *next;
} host_cfg_t;

/* config_t holds a configuration */
typedef struct config_t {
	host_cfg_t 	*host;    /* list of host definitions */
	cmd_def_t 	*command; /* list of command definitions */
	short 		admin_mode; /* global administration mode type */
	unsigned char	admin_digest[SHA1_HASH_SIZE]; /* password hash (calculated) */
	int		max_failures; /* TODO: number of trials for admin mode */
	char		logfile[MAX_LINE_LENGTH]; /* tail this logfile */
	char		*name;        /* FIXME: what's this for? */
} config_t;

/* list of pointers to command strings/digests */
typedef struct cmd_addr_t cmd_addr_t;
struct cmd_addr_t {
	tsw_cmd_t	*addr;
	cmd_addr_t	*next;
};

/* state_t holds a list of open terminals */
typedef struct state_t {
	char		*host;	/* hostname */
	int		pty;	/* pseudo terminal number */
	short		in_sequence; /* expect the next cmd in sequence? */
	unsigned char	*next_expected_hash;
	int		failures;
	int		automated_action; /* are we within an automated action */
	long		hashes;		  /* number of hashes so far */
	cmd_addr_t	*cmd_addr; /* list of pointers to commands/hashes */
	short		maintenance; /* state in maintenance? */
	struct state_t	*next;
} state_t;

/* cmd_line_t holds parsed cmd line args */
struct cmd_line_t {
	char		*file; /* configuration file */
	int		debug; /* debug level */
	int		s;     /* flag for creating a sha1 hash */
	int		q;     /* flag for quiet mode */
};


int lex_state;

/********************* */
/* function prototypes */
/********************* */

/* used for tailing on the logfile */
int open_log(struct file_spec *s_file);
int proc_buf(struct state_t *state, char *buf, int bytes_read);
int tail(struct state_t *state, file_spec s_file); 
long dump_remainder (struct state_t *state, int fd, off_t n_bytes);
int tail_on_file(struct config_t *c, struct state_t *state);

/* used to parse logfile lines */
int proc_line(struct state_t *state, char *line, int length);
int check_digest_validity(struct state_t *state, char *hash, char *host, int pty);
int pty_hash(struct state_t *state, const char *line, int length);
int find_host(const char *line, int length, char *host, int dest_length);
int find_tsw_in_line(const char *line, char *hash, int hash_length, int *tty, long *linenumber);

/* various helper functions */
/* (aren't they *all* helper functions?!) */
void dprintf(int l, char *fmt, ...);
void setprogname(char *str);
char *estrdup(char *s);
char * progname(void);
void eprintf(char *fmt, ...);
void dprintf(int l, char *fmt, ...);

/* (de-)allocate data structures */
int free_host_cfg(host_cfg_t *host);
int free_cmd_def(cmd_def_t *command_def);
int free_all(config_t *c);

struct cmd_def_t *new_cmd_def(void);
struct tsw_cmd_t *new_cmd(void);
struct tsw_time_t *new_time(void);
struct cmd_def_t * add_cmd_def_l(cmd_def_t *listp, cmd_def_t *newp);
struct tsw_cmd_t * add_cmd_l(tsw_cmd_t *listp, tsw_cmd_t *newp);
struct tsw_time_t * add_time_front(tsw_time_t *listp, tsw_time_t *newp);
struct host_cfg_t *addfront(host_cfg_t *listp, host_cfg_t *newp);
void *ecalloc(size_t nelem, size_t elsize);
struct host_cfg_t *new_host_cfg(void);
struct host_cfg_t *get_host(host_cfg_t *host, char *name);
int add_cmd(config_t *c, char *s, int mode);

int hex2bin_hash(char *s, unsigned char *digest);
int hex2bin(char h1, char h2);
int cleanup_hex(char *s1, char *s2);

struct state_t *add_state(state_t *s, char *host, int pty);
struct state_t *rem_state(state_t *state, char *host, int pty);
struct state_t *add_state_f(struct state_t *listp, struct state_t *newp);
struct state_t *new_state(void);
struct state_t *get_state(state_t *s, char *host, int pty);

struct cmd_addr_t *new_cmd_addr(void);
struct cmd_addr_t *add_cmd_addr_front(cmd_addr_t *listp, cmd_addr_t *newp);

se_time_t	*new_se_time(void);
se_time_t	*add_se_time_front(se_time_t *listp, se_time_t *newp);

int add_start(tsw_time_t *tsw_time, int day, se_time_t *se);
int add_end(tsw_time_t *time_spec, int day, int hour, int minute, int second);

/* various find and value check functions for our data structures */
int check_maintenance(config_t *config, state_t *s, char *hash);
int check_hash(config_t *config, state_t *s, char *hash);
struct cmd_def_t *find_cmd_def_by_host(config_t *config, char *host);
struct cmd_def_t *find_cmd_def_by_name(config_t *config, char *command);
cmd_info_t *find_cmd_info(config_t *config, state_t *state, char *hash);
cmd_info_t *find_hash(tsw_cmd_t *allowed_cmd_set, cmd_def_t *cmd_def, char *hash, state_t *state);
cmd_info_t *find_hash_helper(cmd_def_t *command_list, state_t *state, char *hash);

int cmp_hash(const char *asc_hash, const unsigned char *bin_hash);
int time_ok(tsw_time_t *tsw_time);
int check_cmd_addr(cmd_addr_t *cmd_addr, tsw_cmd_t *command);

/* state update */
int update_state_cmd(state_t *state, tsw_cmd_t *command);

/* functions to preprocess configuration file data */
int preprocess_time_spec(config_t *config);
int preprocess_syslog_spec(config_t *config);
int preprocess_time(tsw_time_t *time_spec, int day);
long calc_delta(tsw_time_t *time_spec); 

/* maintenance mode related */
int maint_pass(config_t *config);
char *read_pass(const char *s);
int gen_pass(void);

/* 'intruder found' functions */
int automated_action(state_t *state, host_cfg_t *host_cfg);
int log_alert(host_cfg_t *host_cfg);
int log_automated_action(host_cfg_t *host_cfg);
int log_ok(host_cfg_t *host_cfg);

/* cmd line args */
int read_config(struct cmd_line_t *cmd_line_opts);
struct cmd_line_t *cmd_parse(int argc, char **argv);
void usage(void);

#endif	/* _C_TSW_H */
